/*
 *  Alpha-Mini CCD Cam Driver
 *
 *  Copyright (C) 2000 Florin Boariu.
 *	
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Created: February 2000
 */

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <cgi.h>
#include <syslog.h>
#include <ctype.h>
#include <signal.h>

/* own stuff */
#include <pgtools.h>

#define CAMDEV "/dev/ccdcam"
#define CAMSEM "/tmp/camsem"

#define PICDATA    "/tmp/image.raw16"
#define PICWIDTH   "/tmp/image.width"
#define PICHEIGHT  "/tmp/image.height"

int  devopen (const char *name);                             /* open device */
void camreset (int camfd);                                     /* reset cam */
int  camsemchk (char *buf, int *max);                    /* check semaphore */
int  camsemset (int secsm, int w, int h, const char *msg); /* set semaphore */
void imglite (int camfd, int *width, int *height, int *secs); /* lite image */
void imgread (int camfd, char *buf, int max);                 /* read image */
int  stoi (const char *str);                      /* enhanced string to int */
void segv (int signum);

int main (int argc, char **argv)
{
	int camfd;
	int width, height, secs;
	int len;
	char buf[1000];
	const char *tmp;

	/* http/html head */
	pgopen ("Luigi's pizza is best pizza!", "CCD Cam Request Results");
	if (!cgi_init()) {
		senderr ("cgi_init: %s", cgi_strerror (cgi_errno));
		pgclose ();
		return -1;
	}

	signal (SIGSEGV, segv);

	len = 999;
	if (camsemchk (buf, &len)) {
		buf[len]=0;
		senderr ("camreq: The camera is already in use:\n%s\n",
			    buf);
		pgclose();
		return -1;
	}
	
	camfd = devopen (CAMDEV);

	tmp = cgi_param ("reset");	
	if  (tmp)
		if (!strcmp(tmp, "yes"))  camreset (camfd);
		
	imglite (camfd, &width, &height, &secs);

	if (!camsemset (secs-2, width, height, "eat my shorts!")) {
		senderr ("camreq: could not set semaphore: %s\n",
			 strerror (errno));
	}

	printf ("Congratulations!<br>\n"
		"You should be able to download the image from "
		"<a href=\"/cgi-bin/mkpg?body=jpeg.body&title=Alpha+Mini&head=JPEG+Download\">here</a> "
		"in about %d seconds.\n<br>"
		"Estimated size: %d octets.",
		secs+5, width*height*2);

	close (camfd);
	cgi_done();

	pgclose ();

	return 0;
}

void segv (int num)
{
	fflush (stdout);
	fflush (stderr);
	printf ("\n<br><b>Access violation</b> (a.k.a. segmentation fault, signal %d)"
		" within the readcam-script.<br>"
		"This definitely is a bug. Please submit a full "
		"bug report.", num);
	fflush (stdout);
	exit (0);
}

/*
 * similar to atoi, except for it does filter
 * all numbers out of 'str' and converts them to
 * ints. example
 * "2 is balls is 2 balls too much when 4 angry horses kick you"
 * would be converted to the integer 224.
 *
 * if 'str' contains any "-", the int will be -= -1;
 * ("-23" == -23, "23-" == -23, "2-3" == -23, "-2-3" == 23)
 *
 * the point of this function is to allow the use of
 * the <option> - Tag in HTML files, like in "Frame 1"
 * or "binning of 4 pixels" and make them look good
 *
 * the default return value is 0 (if no numbers found).
 */

int stoi (const char *str)
{
        int nr=0;
	char mul=1;

        if (!str) return 0;
        
        for (; *str; str++)
		if (*str >= '0' && *str <= '9') {
			nr *= 10;
			nr += (*str)-(int)'0';
		} else if (*str == '-') mul *= -1;
        
        return nr*mul;
       
}

int camsemset (int secs, int w, int h, const char *msg)
{
	int fd;
	char buf[1000];
	mode_t m = umask (0);

	if (-1 == (fd = open (CAMSEM, O_RDWR | O_CREAT, 0666))) {
		printf ("open error\n");
		return 0;
	}
	
	umask (m);
	
	snprintf (buf, 999, "time=%d\nmsg=\"%s\"\nw=%d\nh=%d\ndev=%s\n",
		 secs, msg, w, h, CAMDEV);
	if (strlen (buf) > write (fd, buf, strlen (buf))) {
		printf ("write error\n");
		return 0;
	}
	
	close (fd);

	return 1;
}

/*
 * will check for /tmp/camsem. If it exists, some other
 * guy already uses the cam, so it will fail (and return the
 * contents of /tmp/cam_semaphore, which ideally will contain
 * some information about who it is and how long will it take)
 *
 * If "force=yes" is specified, the existing semaphore file
 * will be deleted (and the semaphore process will die by itself)
 */
int camsemchk (char *buf, int *max)
{
	int force=0, fd, r;
	const char *tmp = cgi_param ("force");
	force = tmp ? !strcmp (tmp, "yes") : 0;

	fd = open (CAMSEM, O_RDONLY);

	if (fd > 0) {
		if (!force) {
			r = read (fd, buf, *max);
			*max = r;
			close (fd);
			return 1;
		} else {
			close (fd);
			unlink (CAMSEM);
			
			/* delete old pic -- UGLY HACK!! */
			unlink (PICDATA);
			unlink (PICWIDTH);
			unling (PICHEIGHT);
			return 0;
		}
	} else close (fd);

	/* delete old pic -- UGLY HACK!! */
	unlink (PICDATA);
	unlink (PICWIDTH);
	unling (PICHEIGHT);
	
	return 0;
}

/* 
 * reads all setup-related CGI params, inits the cam
 * and lites the frame
 *
 * it passes back the width and height of the cam, and
 * the liting time in seconds
 */
void imglite (int camfd, int *width, int *height, int *secs)
{
#define _TOP 0
#define _LEFT 1
#define _WIDTH 2
#define _HEIGHT 3
#define _XBIN 4
#define _YBIN 5
#define _TIME 6
	static char params[10][7]={"top", "left", "width", "height", 
				   "xbin", "ybin", "time"};
	char buffer[500];
	int val, i;
	const char *tmp;
	char tmp2[50];
	int len;

	inline int h2c (int bin)
		{
			switch (bin) {
			case 0:
			case 1:
			default:
				return 0x00;
			case 2:
				return 0x01;
			case 4:
				return 0x02;
			case 8:
				return 0x03;
			}
		};

	strcpy (buffer, "init-cam:");
	for (i = 0; i < 7; i++) {
		tmp = cgi_param (params[i]);
		val = tmp ? stoi (tmp) : -1;
		if (val > -1) {
			
			switch (i) {
			case _TOP:
				if (val > 290) sendwarn ("camreq: top ought to be <= 290\n");
				if (val < 0) sendwarn ("camreq: top ought to be >= 0\n");
				break;
			case _LEFT:
				if (val > 784) sendwarn ("camreq: left ought to be <= 784\n");
				if (val < 0) sendwarn ("camreq: left ought to be >= 0\n");
				break;
			case _WIDTH:
				if (val > 784) sendwarn ("camreq: width ought to be <= 784\n");
				if (val < 0) sendwarn ("camreq: width ought to be >= 0\n");
				if (val % 8) sendwarn ("camreq: width ought to be a multiple of 8\n");
			  *width = val;
			  break;
			case _HEIGHT:
				if (val > 290) sendwarn ("camreq: height ought to be <= 290\n");
				if (val < 0) sendwarn ("camreq: height ought to be >= 0\n");
				
				*height = val;
				break;
			case _XBIN:    *width /= (val?val:1); val=h2c(val); break;
			case _YBIN:
				if (*width % (val?val:1))
					sendwarn ("camreq: height ought to be a multiple "
						  "of the y-bin value%d\n",
						  val?val:1);
				*height /= (val?val:1); val=h2c(val);
				break;

			case _TIME:
				if (val > 10000) {
					sendwarn ("camreq: this is a test version, so the integration time "
						  "is limited to 10 seconds.\n");
					val = 10000;
				}
				*secs = val/1000;
				break;
			}
			sprintf (tmp2, "%s=%d,", 
				 params[i], val);
		}
		strcat (buffer, tmp2);
	}

	len = strlen (buffer);
	if (len > write (camfd, buffer, len)) {
		senderr ("camreq: I/O error while init-cam. Maybe timeout?\n");
		pgclose ();
		exit (0);
	}
	
	val = stoi (cgi_param ("frame"));
	sprintf (buffer, "lite-frame:frame=%d", val ? val : 1);
	len = strlen (buffer);
	if (len > write (camfd, buffer, len)) {
		senderr ("camreq: I/O error while lite-frame. Maybe timeout?\n");
		pgclose ();
		exit (0);
	}

#undef _TOP 0
#undef _LEFT 1
#undef _WIDTH 2
#undef _HEIGHT 3
#undef _XBIN 4
#undef _YBIN 5
#undef _TIME 6
}

void camreset (int camfd)
{
	char reset_str[]="reset-cam:hitime=10,lotime=30";
	write (camfd, reset_str, strlen (reset_str));
}

int devopen (const char *name)
{
	int camfd = open ("/dev/ccdcam", O_RDWR);
	if (camfd < 1) {
		senderr ("camreq: could not access cam: %s",
			    strerror (errno));
		pgclose ();
		exit (-1);
	}

	return camfd;
}






