#ifndef __AMIN_H
#define __AMIN_H

/*
 *  Alpha-Mini CCD Cam Driver
 *
 *  Copyright (C) 2000 Florin Boariu.
 *	
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Created: January 2000
 */

#include <linux/kernel.h>

/* driver header files */
#include <linux/fs.h>
#include <linux/wrapper.h>

#include <linux/string.h>


/* Alpha-Mini Commands (from kernel to hardware) */
#define AMIN_NOP       (0)
#define AMIN_SENDLINE  (1)
#define AMIN_FRAME1    (2)
#define AMIN_FRAME2    (3)
#define AMIN_MOVELINE  (4)
#define AMIN_INIT      (5)


/* ioctl numbers (commands from user to kernel) */

/* hardware interaction commands */
#define AMIN_RESET_CAM   0x601  
#define AMIN_UNRESET_CAM 0x602
#define AMIN_INIT_CAM    0x603
#define AMIN_UNINIT_CAM  0x604
#define AMIN_LITE_FRAME  0x605
#define AMIN_RECEIVE     0x606
#define AMIN_SYNC        0x607
#define AMIN_NO_OP       0x607

/* cam setup commands */
#define AMIN_SET_BINNING 0x610
#define AMIN_SET_REGION  0x611
#define AMIN_SET_TIME    0x612
#define AMIN_GET_WIDTH   0x613
#define AMNI_GET_HEIGHT  0x614
#define AMIN_GET_TIME    0x615

/* for debugging purposes only */
#define AMIN_PRINT_INFO  0x620


/* physical size of the image (optical cells in the cam) */
#define AMIN_PHYS_WIDTH  784
#define AMIN_PHYS_HEIGHT 290

/* requested version of the hardware */
#define AMIN_REQ_VERSION 3

/* minor numbers */
#define AMIN_MIN_RAW8    0
#define AMIN_MIN_RAW12   1
#define AMIN_MIN_RAW16   1
#define AMIN_MIN_SETUP   2
#define AMIN_MIN_STATE   3

/*
 * setup data of the cam
 */
typedef struct _amin_cam_setup {
	int ystart, ysize;      /* physical height of the logical image (no binning, get that? ;-) */
	                        /* the logical width will be computed (=> binning, areas...) */
	unsigned char xbin;     /* horizontal binning */
	unsigned char ybin;     /* vertical binning */
	unsigned char left;     /* left area/8  (thrown away) */
	unsigned char mid;      /* mid area/8   (kept) */
	unsigned char right;    /* right area/8 (thrown away) */
	unsigned char version;  /* hardware version */
	unsigned long int time; /* liting time in msecs */
}amin_cam_setup;

/*
 * status of the cam (what has done, what hasn't -- some kind
 *                    of dependency data base)
 */
typedef struct _amin_cam_state {
	char reset;       /* 0: no reset occured, 1: reset did occur */
	char init;        /* 0: no init occured,  1: init did occur */
	char lite;        /* 0: no lite           1: lite */
	char frame;       /* 0: no lite; 1: frame 1 lited; 2: frame 2 lited */
	char hclip;       /* 0: no h-clipping done; 1: clipping done (first lines removed) */
	int maxX, maxY;   /* real width and height of the image in logical pixels */
	int lastX, lastY; /* pixels that have already been read */
}amin_cam_state;

typedef struct _amin_cam_info {
	struct _amin_cam_setup setup;
	struct _amin_cam_state state;
} amin_cam_info;

typedef struct _amin_kbuf {
	char buf[1024];
	int len, pos;
} amin_kbuf;

/* fops functions */

static int      amin_open (struct inode*, struct file*);
static int      amin_ioctl (struct inode*, struct file*, unsigned int cmd, void *arg);

static int      amin_release (struct inode *inode, struct file *file);
static ssize_t  amin_read (struct file *file, char *buffer, size_t count, loff_t *offset);
static ssize_t  amin_write (struct file *file, char *buffer, size_t count, loff_t *offset);

static int amin_read_img (char *buf, int len, int bpp);
static int amin_read_kbuf (char *buf, int len, amin_kbuf*);

void amin_mksetupbuf (amin_cam_setup*, amin_kbuf*);
void amin_mkstatebuf (amin_cam_state*, amin_kbuf*);

void amin_sfcpy (char *where, const char *fmt, ...);

/* parse helpers */
int next_fs_block (char **fsbuf, int *len, char *next, int nextlen);
int strtoint (const char *str);
inline const char *findval (const char *var, const char *key);           /* returns the value */
inline long int getint (const char *text, const char *key, long int def);

/* cam control */

/* sends RESET to the cam */
void amin_reset_cam (amin_cam_state *stat, int lo_time, int hi_time); 

/* 
 * sets state.reset to ZERO, so that on next LITE, a
 * RESET will occur
 */
void amin_unreset_cam (amin_cam_state *stat);

/* sets the binning in the SETUP-Struct */
void amin_set_binning (amin_cam_info *info, int xbin, int ybin);

/* sets the logical size of the image */
void amin_set_region (amin_cam_info *info, int x, int y, int w, int h); 

/* sets the liting time */
void amin_set_time (amin_cam_info *info, unsigned long int msecs);

/* sends INIT to the cam */
int amin_init_cam (amin_cam_info *info);

void amin_uninit_cam (amin_cam_state *stat);

/* lites the respective frame */
int amin_lite_frame (amin_cam_info *info, int frame);

void amin_check_setup (amin_cam_setup *setup);

void amin_print_info (amin_cam_info *info);

inline void amin_sync (unsigned long msec);

 /* 
  * sends a NOP to the cam -- mostly for sync purposes 
  * (wait for liting, e.g.) CAREFUL! It might block for
  * a long time (until the liting time is up) or forever
  * (if the cam cannot answer anymore, for example if it
  * has not been reset or if it's been damaged/removed).
  */
/* void amin_nop (void); */
/*
 * PS: We don't need it -- we'll simply put a SEND_BYTE_W (AMIN_NOP), that's short enough
 */

#endif /* __AMIN_H */
