#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <linux/stepeng_ioctl.h>
#include <math.h>

#include <drivemsg.h>

#include <astro/coord.h>

#define CONFIG_MAX 16384
#define MAX_RUN_TIME 1000.0

#define OPT_AZMDEV 'a' /* azimuth control device */
#define OPT_ELVDEV 'e' /* elevation control device */
#define OPT_DAEMON 'd' /* start in daemon mode */
#define OPT_CONFIG 'c' /* calibration data: x,y,z */
#define OPT_MSGKEY 'k' /* message queue key */
#define OPT_HELP   'h' /* show help */

void show_help_drived (const char *name)
{
	fprintf (stderr, "\
Usage: %s OPTIONS\n\
\n\
Valid OPTIONS are:\n\
\n\
     -a,--azmdev dev    Azimuth control device\n\
     -c,--config file   Config file\n\
     -d,--daemon        Fork into background (daemon mode)\n\
     -e,--elvdev dev    Elevation control device\n\
     -k,--msgkey key    Message queue key\n\
     -h,--help          Shows this screen.\n\
\n\
Drive Control Daemon.\n\
\n", name);
}

/*
 * data about one engine (hardware
 */
#define ENG_STOPPED 0
#define ENG_RUNNING 1
struct engine_info {
	double speed;   /* the last set speed */
	double tstart;  /* when was the speed set */
	double tend;    /* when is it supposed to stop */
	char stat;	
};

/*
 * data about one engine (software, translations)
 */
struct engine_device {
	struct engine_info einfo; /* information about engine settings */
	char dname[256];          /* device name */
	int fd;                   /* file descriptor */
	double transl;            /* translation factor
				   * (telescope angle to engine angle) */
	double zero;              /* the initial engine position
				   * (zero degree movement) */
	double maxspeed;          /* the engine's maximal speed */
	double lastecheck;        /*
				   * the last time chkengine
				   * was called -- simple benchmark
				   * since this function is called
				   * more or less periodically,
				   * it gives us a vague idea
				   * about when it's going to be
				   * called next time. this way we
				   * can approximate the
				   * speed to set a little better
				   * (we must only use time
				   * slices of 'echeckfreq' (the
				   * check frequency).
				   */
	double echkfreq;
};

/*
 * main application data (devices, settings...)
 */
struct teleinfo {
	struct engine_device dax1, dax2;
	double clx, cly, clz;
	/*double azmplus, elvplus; *//* azimuth and elevation correction */
	double rollangle, tiltangle, turnangle;  /* telescope position correction */
	int qid;
	struct hpe_accel_ramp ramp;
	char config[256], rampfile[256];
	double a1min, a1max, a1pos; /* the mechanics constraints (axis 1 and axis 2) */
	double a2min, a2max, a2pos;
	int a1loaction, a1hiaction;
	int a2loaction, a2hiaction;
	double ggphh;  /* geographic height */
};

/* application stuff -- configuration, devices */
struct teleinfo *tele_open (const char *azm, const char *elv,
			    const char *config);
void tele_close (struct teleinfo *tele);
int dev_open (struct engine_device *eng);
void dev_close (struct engine_device *eng);
int tele_serve (struct teleinfo *tele, struct telecmd *cmd);

/* high level functions (movement logics) */
int tele_moveto (struct teleinfo *tele, struct c_cart *coord, double time); /* moving functions */
int tele_setspeed (struct teleinfo *tele, double azm, double elv, double time);
int tele_chkengine (struct engine_device *dev);

/* low level functions (movement physics) */
int eng_setspeed (struct engine_device *dev, double speed, double till_when);
void eng_showstat (struct engine_device *dev);
int eng_moveby (struct teleinfo *tele, double byx1, double byx2, double till);

inline void coord_to_axis (struct teleinfo*, struct c_polar *coord,
			   double *axis1, double *axis2);
/*inline void axis_to_coord (double axis1, double axis2,
			   struct c_polar *coord);*/
			   
void mkramp (struct hpe_accel_ramp *ramp, int size, double max_t, double max_sp);
int load_ramp (struct hpe_accel_ramp *ramp, const char *file);
void atm_correct (struct c_polar *coord);

int main (int argc, char **argv)
{
	struct option opts[]={{"azmdev",   1, NULL, OPT_AZMDEV},
			      {"elvdev",   1, NULL, OPT_ELVDEV},
			      {"daemon",   0, NULL, OPT_DAEMON},
			      {"config",   1, NULL, OPT_CONFIG},
			      {"msgkey",   1, NULL, OPT_MSGKEY},
			      {"help",   0, NULL, OPT_HELP}};
	char azmdev[256], elvdev[256];
	double clx=0.0, cly=0.0, clz=0.0;
	char daemon=0;
	int opt;
	const char *tmp;
	int key=0;
	struct teleinfo *tele;
	struct message msg;
	char run=1;
	char cfgfile[256];

	azmdev[0] = 0;
	elvdev[0] = 0;

	while (0 < (opt=getopt_long(argc, argv, "a:c:de:hk:", opts, NULL))) {
		switch (opt) {
		case OPT_AZMDEV:
			strncpy (azmdev, optarg, 255);
			azmdev[255]=0;
			break;
			
		case OPT_ELVDEV:
			strncpy (elvdev, optarg, 255);
			elvdev[255]=0;		
			break;

		case OPT_DAEMON:
			daemon = 1;
			break;

		case OPT_MSGKEY:
			key = atoi (optarg);
			break;
			
		case OPT_CONFIG:
			strncpy (cfgfile, optarg, 255);
			cfgfile[255]=0;
			break;

		case OPT_HELP:
			show_help_drived (argv[0]);
			exit(0);
			break;

		default:
			break;
		}
	}

	if (daemon) switch (fork()) {
	case -1:
		perror ("fork");
		return -1;

	case 0:
		printf ("drvctl forking into background\n");
		return 0;

	default:
		break;
	}

	tele = tele_open (azmdev[0] ? azmdev : NULL,
			  elvdev[0] ? elvdev : NULL,
			  cfgfile);
	if (!tele) return 0;

	if (-1 == (tele->qid = msgget (key, 0622 | IPC_CREAT))) {
		perror ("msgget");
		return 0;
	}

	printf ("queue id: %d\nqueue key: %d\n", tele->qid, key);

	do {
		bzero (&msg, sizeof (struct message));
		if (-1 != msgrcv (tele->qid, (struct msgbuf*)&msg, 
				  sizeof (struct telecmd), MSG_TELE,
				  IPC_NOWAIT)) {

			printf ("got message 0x%x\n", msg.cmd.cmd);
			if (msg.cmd.cmd == DAEMON_QUIT) run = 0;
			else tele_serve (tele, &(msg.cmd));

		} else if (errno != ENOMSG) {
			perror ("msgrcv");
			run = 0;
		} else usleep (1000);

		tele_chkengine (&(tele->dax1)); /* stop the engines if we reached the given time */
		tele_chkengine (&(tele->dax2));

	} while (run);

	tele_close (tele);

	if (-1 == (msgctl (tele->qid, IPC_RMID, NULL))) {
		perror ("msgctl(IPC_RMID)");
		return 0;
	}

	return 0;
}

int tele_chkengine (struct engine_device *eng)
{
	double now = get_time();
	double nospeed = 0.0;
	eng->echkfreq = (eng->echkfreq+(now-eng->lastecheck)) / 2.0;
	eng->lastecheck = now;
	if (now >= eng->einfo.tend) {
		if (eng->einfo.stat == ENG_RUNNING) {	
			printf ("default stop\n");
			if (-1 == ioctl (eng->fd, HPE_TOSPEED, &nospeed))
				perror ("tele_chkengine");
			eng->einfo.stat = ENG_STOPPED;
		}
	}
}

void mkramp (struct hpe_accel_ramp *ramp, int size, double max_t, double max_sp)
{
	int i;
	double t = 0.0;
	double speed;
	ramp->size = size;
	ramp->tslice = max_t/(double)size;

	for (i = 0; i < size; i++,t+=ramp->tslice) {
		speed = max_sp*(1.0-1.0/exp(t));
		ramp->ramp[i].speed = speed;
		ramp->ramp[i].dist = ramp->tslice*speed;
		printf ("adding to accel ramp: %.5f %.5f %.5f\n",
			t, ramp->ramp[i].speed, ramp->ramp[i].dist);
	}
}

double extract_dbl (const char *text, const char *key, double def)
{
	char *pos = strstr (text, key);
	if (pos) return atof(pos+1+strlen (key));
	else return def;
}

int extract_str (const char *text, const char *key, char *buf, int bufsize)
{
	char *pos = strstr (text, key);
	int i=0;
	if (pos) {
		pos += strlen (key) + 1;
		while (i < bufsize-1 && *pos && isgraph(*pos)) {
			buf[i++] = *pos++;
		}
		buf[i] = 0;
	} else {
		printf ("%s not found in:\n%s\n", key, text);
		*buf = 0;
		return 0;
	}
	return i;
}

struct teleinfo *tele_open (const char *azm, const char *elv,
			    const char *config)
{
	struct teleinfo *tele = (struct teleinfo*)
		malloc(sizeof(struct teleinfo));
	int cfg = config ? open (config, O_RDONLY) : -1;
	char cfgtext[CONFIG_MAX];
	int cfgsize=0, rb=0;
	double fac;

	strncpy (tele->config, config, 255);
	tele->config[255] = 0;
	
	if (!tele) return NULL;	

	if (cfg > -1) {      
		/* read config file, set params */
		cfgsize = lseek (cfg, 0, SEEK_END);
		 lseek (cfg, 0, SEEK_SET);
	} else cfgsize = 0;

	if (cfgsize > CONFIG_MAX-1) {
		fprintf (stderr, "Config file too big. Maximum size: %d\n",
			 CONFIG_MAX-1);
		exit (-1);
	}

	if (cfg) {
		printf ("reading %d bytes from %s\n", cfgsize, config);
		fflush (stdout);
		rb = read (cfg, cfgtext, cfgsize);
	        printf ("read %d bytes from %s\n",
			rb, config);
		fflush (stdout);
		if (rb < 0) {
			perror (config);
			rb = 0;
		}
		cfgtext[rb] = 0;
	} else {
		printf ("config file not found\n");
		cfgtext[cfgsize] = 0;
	}

	/* general settings -- device names */
	if (azm) strncpy (tele->dax1.dname, azm, 255);
	else extract_str (cfgtext, "axis1-device", tele->dax1.dname, 255);
	dev_open (&(tele->dax1));

	if (elv) strncpy (tele->dax2.dname, elv, 255);
	else extract_str (cfgtext, "axis2-device", tele->dax2.dname, 255);
	dev_open (&(tele->dax2));

	/* calibration stuff, translations, mechanics... */
	tele->dax1.transl = extract_dbl (cfgtext, "axis1-translation", 1.0);
	tele->dax2.transl = extract_dbl (cfgtext, "axis2-translation", 1.0);
	tele->rollangle   = extract_dbl (cfgtext, "roll-angle", 0.0);
	tele->tiltangle   = extract_dbl (cfgtext, "tilt-angle", 0.0);
	tele->dax1.zero   = extract_dbl (cfgtext, "axis1-zero", 0.0);
	tele->dax2.zero   = extract_dbl (cfgtext, "exis2-zero", 0.0);
	tele->ggphh       = extract_dbl (cfgtext, "latitude", 0.0);

	/* this will get complicated later */
	tele->a1min       = extract_dbl (cfgtext, "axis1-mimimum", 0.0);
	tele->a1max       = extract_dbl (cfgtext, "axis1-maximum", 0.0);
	tele->a1pos       = extract_dbl (cfgtext, "axis1-current", 0.0);
	tele->a2min       = extract_dbl (cfgtext, "axis2-mimimum", 0.0);
	tele->a2max       = extract_dbl (cfgtext, "axis2-maximum", 0.0);
	tele->a2pos       = extract_dbl (cfgtext, "axis2-current", 0.0);		

	if (0 < extract_str (cfgtext, "acceleration-ramp", tele->rampfile, 255)) {
		printf ("reading acceleration ramp from %s\n", tele->rampfile);
		if (-1 == load_ramp (&(tele->ramp), tele->rampfile)) {
			perror (tele->rampfile);
			exit (-1);
		}
		if (-1 == ioctl (tele->dax1.fd, HPE_SETRAMP, &(tele->ramp)))
			perror ("HPE_SETRAMP(ax1)");
		if (-1 == ioctl (tele->dax2.fd, HPE_SETRAMP, &(tele->ramp)))
			perror ("HPE_SETRAMP(ax2)");
	} else {		
		printf ("using default acceleration ramp (bad idea)\n");
		if (-1 == ioctl (tele->dax1.fd, HPE_GETRAMP, &(tele->ramp)))
			perror ("HPE_GETRAMP(ax1)");
	}

	printf ("ramp size: %d\ntime slice: %f\nstart speed: %f\nmaximum speed: %f\n",
		tele->ramp.size, tele->ramp.tslice, tele->ramp.ramp[1].speed,
		tele->ramp.ramp[tele->ramp.size-1].speed);

	/*ioctl (tele->dax1.fd, HPE_GETRAMP, &(tele->ramp));*/

	fac = extract_dbl (cfgtext, "engine-speed-translation", 0.0);
	if (fac != 0.0) {
		printf ("setting speed translation factor to %f\n", fac);
		if (-1 == ioctl (tele->dax1.fd, HPE_SETSPDFAC, &fac))
			perror ("HPE_SETSPDFAC(ax1)");	       
		if (-1 == ioctl (tele->dax2.fd, HPE_SETSPDFAC, &fac))
			perror ("HPE_SETSPDFAC(ax2)");
	}

	fac = extract_dbl (cfgtext, "engine-position-translation", 0.0);
	if (fac != 0.0) {
		printf ("setting position translation factor to %f\n", fac);
		if (-1 == ioctl (tele->dax1.fd, HPE_SETPOSFAC, &fac))
			perror ("HPE_SETPOSFAC(ax1)");
		if (-1 == ioctl (tele->dax2.fd, HPE_SETPOSFAC, &fac))
			perror ("HPE_SETPOSFAC(ax2)");
	}

	printf ("axis 1 control status:\n");
	eng_showstat (&(tele->dax1));

	printf ("axis 2 control status:\n");
	eng_showstat (&(tele->dax2));

	close (cfg);
	
	return tele;
}

/*
 * loads an acceleration ramp from file
 * 
 * ramp format: <size> <timeslice> <0.0> <speed2> <speed3> ... <speed size-1>
 */
int load_ramp (struct hpe_accel_ramp *ramp, const char *filename)
{
	int i;
	char s[50];
	FILE *f = fopen (filename, "r");
	if (!f) return -1;

	fscanf (f, "%d", &(ramp->size));
	fscanf (f, "%s", s);
	ramp->tslice = atof(s);

	printf ("ramp items: %d\ntime slice: %f\n",
		ramp->size, ramp->tslice);

	printf ("reading accel ramp");
	for (i=0; i < ramp->size; i++) {
		fscanf (f, "%s", s);
		ramp->ramp[i].speed = atof(s);
		/* printf ("added: %f\n", ramp->ramp[i].speed); */
		printf (".");	       
		/* the integral is missing!! */
	}
	printf ("\n");

	fclose (f);
}

int dev_open (struct engine_device *eng)
{
	if (-1 == (eng->fd = open (eng->dname, O_RDWR))) {
		perror (eng->dname);
		return -1;
	}
	return 0;
}

void dev_close (struct engine_device *eng)
{
	close (eng->fd);
	eng->fd = -1;
}

void eng_showstat (struct engine_device *dev)
{
	struct hpe_engine_state stat;

	if (-1 == ioctl (dev->fd, HPE_RENEWSTATE, &stat))
		perror ("HPE_RENEWSTATE");		
	
	printf ("\tspeed:\t%2.5f\n\ttime:\t%2.5f\n\tpos:\t%2.5f\nlimits:\t%s %s\n",
		stat.speed, stat.time, stat.pos,
		HPE_HILIM(stat.lim) ? "yes" : "no",
		HPE_LOLIM(stat.lim) ? "yes" : "no");
}

void tele_close (struct teleinfo *tele)
{
	FILE *cfg = fopen (tele->config, "w");
	double fac;
	
	fprintf (cfg, "axis1-device %s\n", tele->dax1.dname);	
	fprintf (cfg, "axis2-device %s\n", tele->dax2.dname);
	fprintf (cfg, "axis1-translation %f\n", tele->dax1.transl);
	fprintf (cfg, "axis2-translation %f\n", tele->dax2.transl);
	fprintf (cfg, "roll-angle %f\n", tele->rollangle);
	fprintf (cfg, "tilt-angle %f\n", tele->rollangle);
	fprintf (cfg, "axis1-zero %f\n", tele->dax1.zero);
	fprintf (cfg, "axis2-zero %f\n", tele->dax2.zero);
	fprintf (cfg, "latutide %f\n", tele->ggphh);
	fprintf (cfg, "axis1-minimum %f\n", tele->a1min);
	fprintf (cfg, "axis1-maximum %f\n", tele->a1max);
	fprintf (cfg, "axis1-current %f\n", tele->a1pos);
	fprintf (cfg, "axis2-minimum %f\n", tele->a2min);
	fprintf (cfg, "axis2-maximum %f\n", tele->a2max);
	fprintf (cfg, "axis2-current %f\n", tele->a2pos);	
	if (strlen(tele->rampfile))
		fprintf (cfg, "acceleration-ramp %s\n", tele->rampfile);
	if (tele->dax1.fd > 0) {
		if (-1 == ioctl (tele->dax1.fd, HPE_GETSPDFAC, &fac))
			perror ("HPE_GETSPEEDFAC(ax1)");
		fprintf (cfg, "engine-speed-translation %f\n", fac);
		if (-1 == ioctl (tele->dax1.fd, HPE_GETPOSFAC, &fac))
			perror ("HPE_GETSPEEDFAC(ax2)");
		fprintf (cfg, "engine-position-translation %f\n", fac);
	}

	fclose (cfg);
	
	dev_close (&(tele->dax1));
	dev_close (&(tele->dax2));
	free (tele);
}

int tele_serve (struct teleinfo *tele, struct telecmd *cmd)
{
	struct c_cart coord;
	double t = cmd->data[D_TIME];
	double now = get_time();
	if (t < 0) t = -t;
	if (t < MAX_RUN_TIME*2) t = now+t;
	
	bzero (&coord, sizeof (struct c_polar));
	coord.x = cmd->data[D_X];
	coord.y = cmd->data[D_Y];
	coord.z = cmd->data[D_Z];

	/*printf ("data %.5f %.5f %.5f %.5f makes coordinates %.5f, %.5f, %.5f with message: ",
		cmd->data[0], cmd->data[1], cmd->data[2], cmd->data[3],
		coord.azm, coord.elv, coord.dst);*/

	switch (cmd->cmd) {

		/* first class options -- direct motions */
	case DRV_GOTO: /* go to the specified position */
	  	printf ("GOTO\n");
		tele_moveto (tele, &coord, t);
		break;

	case DRV_SETSPEED: /* set the speed in deg/sec */
	  	printf ("SETSPEED\n");
		tele_setspeed (tele, cmd->data[D_X], cmd->data[D_Y], t);
		break;
		
	case DRV_SHOWSTAT:
		printf ("azimuth control status:\n");
		eng_showstat (&(tele->dax1));
		printf ("elevation control status:\n");
		eng_showstat (&(tele->dax2));
		break;

	default:
		fprintf (stderr, "unknown message: %d %.2f %.2f %.2f %.2f\n",
			 cmd->cmd, cmd->data[0], cmd->data[1], cmd->data[2], cmd->data[3]);
		break;
	}

	return 0;
}

/*
 * high-level movement function.
 * equatorial coordinates, not corrected. time as in "when
 * to reach the specified point"
 */
int tele_moveto (struct teleinfo *tele, struct c_cart *coord, double time)
{
	double milx1, milx2;
	double tox1, tox2;
	double cazm, celv;
	struct c_polar pos, cpol;
	struct c_cart ccor; /* corrected cartesian coordinates */

	/* do the calibration correction: roll<->x, turn<->y, tilt<->z */

	/* x-axis (roll) */
	ccor.y = coord->y * cos(tele->rollangle) - coord->z * sin(tele->rollangle);
	ccor.z = coord->y * sin(tele->rollangle) - coord->z * cos(tele->rollangle);
	ccor.x = coord->x;

	/* y-axis (turn) */
	ccor.z = coord->z * cos(tele->turnangle) - coord->x * sin(tele->turnangle);
	ccor.x = coord->z * sin(tele->turnangle) - coord->x * cos(tele->turnangle);
	ccor.y = coord->y;
	
	/* z-azis (tilt) */
	ccor.x = coord->x * cos(tele->tiltangle) - coord->y * sin(tele->tiltangle);
	ccor.y = coord->x * sin(tele->tiltangle) - coord->y * cos(tele->tiltangle);
	ccor.z = coord->z;
	
	
	polar (ccor.x, ccor.y, ccor.z,
	       &(cpol.dst), &(cpol.elv), &(cpol.azm));

	/* mechanics */
	coord_to_axis (tele, &cpol, &tox1, &tox2);

	if (-1 == ioctl (tele->dax1.fd, HPE_MILEAGE, &milx1))
		perror ("HPE_MILEAGE(ax1)");	      
	if (-1 == ioctl (tele->dax2.fd, HPE_MILEAGE, &milx2))
		perror ("HPE_MLEADE(ax2)");

	return eng_moveby (tele, tox1-milx1, tox2-milx2, time);
}

inline void coord_to_axis (struct teleinfo *tele,
			   struct c_polar *coord,
			   double *axis1, double *axis2)
{
	*axis1 = coord->azm + tele->dax1.zero;
	*axis2 = coord->elv + tele->dax2.zero;
	
	*axis1 -= tele->tiltangle;
	*axis2 -= tele->rollangle;
	
	*axis1 *= tele->dax1.transl;
	*axis2 *= tele->dax2.transl;
}

inline void axis_to_coord (double axis1, double axis2,
			   struct c_polar *coord)
{
	printf ("FIXME! axis_to_coord!\n");
}

int eng_moveby (struct teleinfo *tele, double byx1, double byx2, double till)
{
	double now;
	double tdiff;
	double tx1, tx2, need_t;
	double speed_max = tele->ramp.ramp[tele->ramp.size-1].speed;
	double spdx1, spdx2;

	/* an approximation of the time needed (we actually need more) */
	tx1 = byx1/(byx1 > 0.0 ? speed_max : -speed_max);
	tx2 = byx2/(byx2 > 0.0 ? speed_max : -speed_max);

	need_t = tx1 > tx2 ? tx1 : tx2;
	now = get_time();
	tdiff = till-now;
	if (need_t <= tdiff) need_t = tdiff;       
	spdx1 = byx1/need_t;
	spdx2 = byx2/need_t;

	eng_setspeed (&(tele->dax1), spdx1, now+need_t);
	eng_setspeed (&(tele->dax2), spdx2, now+need_t);
}

/*
 * 
 */
int tele_setspeed (struct teleinfo *tele, double ax1, double ax2, double time)
{
	eng_setspeed (&(tele->dax1), ax1*tele->dax1.transl, time);
	eng_setspeed (&(tele->dax2), ax2*tele->dax2.transl, time);
}

/*
 * atmospheric refraction correction
 */
void atm_correct (struct c_polar *coord)
{
}

int eng_setspeed (struct engine_device *dev, double speed, double till_when)
{
	double now = get_time();
	struct hpe_engine_state stat;
	dev->einfo.tstart = now;
	if (till_when > now+MAX_RUN_TIME || till_when < now) {
		dev->einfo.tend = now+MAX_RUN_TIME;
	}
	else dev->einfo.tend = till_when;

	printf ("engine %s: %f rad/sec for %f seconds\n",
		dev->dname, speed, dev->einfo.tend-now);

	if (-1 == ioctl (dev->fd, HPE_TOSPEED, &speed))
		perror ("HPE_TOSPEED");
	if (-1 == ioctl (dev->fd, HPE_GETSTATE, &stat))
		perror ("HPE_GETSTATE");
	dev->einfo.speed = stat.speed;
	dev->einfo.stat  = ENG_RUNNING;
}








