#ifndef __DRIVEMSG_H
#define __DRIVEMSG_H

#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <stdio.h>

/*
 * Some theory about telescope motions:
 *
 * The Drive Daemon (this) receives motion jobs via the msg
 * IPC facility. These jobs are: GOTO, MOVEBY and SETSPEED.
 * 
 * Some application has the possibility to move the telescope
 * by using one of these commands. The GOTO command expects
 * azm and elv values, aswell as a time value (epoch time,
 * double-precision float) about when the specified position is
 * to be reached. (does the SETSPEED-command make any sense here?)
 * If the time value is 0.0, then "now" is assumed.
 */

/*
 * GOTO takes real 3d point coordinates as arguments.
 * SETSPEED takes the two angle speeds for the physical telescope axis.
 * about MOVEBY... no idea. it doesn't make any sense, does it?
 */

/* first class options -- direct motions */
#define DRV_GOTO     0x11  /* go to the specified position */
#define DRV_SETSPEED 0x13  /* set the speed in deg/sec */

/* other options */
#define DRV_SHOWSTAT 0x40  /* show engine states */

/* daemon control options */
#define DAEMON_QUIT  0xff

#define CMD_CLASS(cmd) (cmd>>5)

struct telecmd {
	unsigned char cmd;
	double data[4];
};

#define MSG_TELE 1
struct message {
	long mtype;
	struct telecmd cmd;
};

#define D_TIME 0
#define D_AZM  1
#define D_ELV  2
#define D_DST  3
#define D_X    1
#define D_Y    2
#define D_Z    3

#define D_LOAZM 0
#define D_LOELV 1
#define D_HIAZM 2
#define D_HIELV 3

#define MKPOLAR(data,coord) coord.azm=data[D_AZM]; coord.elv=data[D_ELV]; coord.dst=data[D_DST]
#define MK3DC(data,coord)   coord.x=data[D_X]; coord.y=data[D_Y]; coord.z=data[D_Z];

inline double get_time(void)
{
	struct timeb t;
	ftime (&t);
	return (double)t.time+(double)t.millitm/1000.0;
}

struct drived {
	int qkey;
	int qid;
};

/* drived communication stuff */

inline struct drived *dd_init (int key)
{
	struct drived *dd = (struct drived *)malloc (sizeof (struct drived));
	dd->qkey = key;
	if (-1 == (dd->qid = msgget (key, 0))) {
		perror ("dd_init");
		free (dd);
		return NULL;
	}
	return dd;
}
inline void dd_close (struct drived *dd)
{
	free(dd);
}

inline int dd_msg (struct drived *dd, int message,
		   double a, double b, double c, double d)
{
	struct message msg;
	msg.mtype = MSG_TELE;
	msg.cmd.cmd = message;
	msg.cmd.data[0] = a;
	msg.cmd.data[1] = b;
	msg.cmd.data[2] = c;
	msg.cmd.data[3] = d;
	return msgsnd (dd->qid, (struct msgbuf*)&msg,
		       sizeof (struct telecmd), 0);
}

#define dd_goto(dd,t,x,y,z)     dd_msg(dd,DRV_GOTO,t,x,y,z)
#define dd_setspeed(dd,t,x1,x2) dd_msg(dd,DRV_SETSPEED,t,x1,x2,0.0)

#endif /* __DRIVEMSG_H */





