#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include <network.h>

#include <drivemsg.h>

#define OPT_DAEMON 'd' /* start in daemon mode */
#define OPT_MSGKEY 'k' /* message queue key */
#define OPT_PORT   'p' /* port to listen on */
#define OPT_STREAM 's' /* turns on streaming mode. default is off */
#define OPT_HELP   'h' /* show help */

void show_help_drvctl (const char *name)
{
	fprintf (stderr, "\
Usage: %s OPTIONS\n\
\n\
Valid OPTIONS are:\n\
\n\
     -d,--daemon        Fork into background (daemon mode)\n\
     -k,--msgkey key    Message queue key\n\
     -p,--port port     Listen on port 'port'\n\
     -s,--stream        Don't read all values at once, have them streamed\n\
     -h,--help          Shows this screen.\n\
\n\
Joystick Control Daemon.\n\
\n", name);
}

typedef struct _eph_t {
	double time;
	double azm;
	double elv;
	double dst;
} eph_t;

/*
 * saves information about a job
 */ 
struct telejob {
	eph_t *ephem;
	unsigned int ephsize;
	unsigned int ephpos;
	double tstart;
	int fd;     /* where to read ehpemerides from in case of streaming */
	char strm;  /* stream 0 no 1 yes */
};

int getdbl (int fd, double *nr);

struct telejob *init_telejob (int fd, char strm);
void close_telejob (struct telejob *tj);
int get_eph (struct telejob *tj, eph_t *e);

int main (int argc, char **argv)
{
	struct option opts[]={{"daemon",   0, NULL, OPT_DAEMON},
			      {"msgkey",   1, NULL, OPT_MSGKEY},
			      {"port",     1, NULL, OPT_PORT},
			      {"stream",   0, NULL, OPT_STREAM},
			      {"help",     0, NULL, OPT_HELP}};
	char daemon=0;
	int opt, sockfd, connfd, port=12345, key;
	struct message msg;
	char run=1, strm=0;
	struct drived *dd;
	struct telejob *tj;
	eph_t e;

	while (0 < (opt=getopt_long(argc, argv, "a:c:de:hk:sp:", opts, NULL))) {
		switch (opt) {
		case OPT_DAEMON:
			daemon = 1;
			break;
			
		case OPT_MSGKEY:
			key = atoi(optarg);
			break;

		case OPT_PORT:
			port = atoi(optarg);
			break;

		case OPT_STREAM:
			strm = 1;
			break;

		case OPT_HELP:
			show_help_drvctl (argv[0]);
			exit(0);
			break;

		default:
			break;
		}
	}

	if (daemon) switch (fork()) {
	case -1:
		perror ("fork");
		return -1;

	case 0:
		printf ("drvctl forking into background\n");
		return 0;

	default:
		break;
	}

	dd = dd_init (key);
	if (!dd) {
		perror("dd_init");
		exit (-1);
	}

	if (-1 == (sockfd = startserv (port, 1))) exit (-1);	

 loop:
	do {
		printf ("waiting for client\n");
		connfd = getconn (sockfd);		
		tj = init_telejob (connfd, strm);
		if (tj) {
			while (-1 != get_eph(tj, &e)) {
				while (get_time() < e.time) usleep (10000);
				/*dd_moveby (dd, e.time, e.azm, e.elv, e.dst);*/
				printf ("%f %f %f\n",
					e.time, e.azm, e.elv);
				dd_setspeed (dd, e.time, e.azm, e.elv, e.dst);
			}
			close_telejob (tj);
		}
		printf ("connection closed\n");
	} while (run);

	dd_close (dd);
	return 0;
}

int getdbl (int fd, double *nr)
{
	char c=' ';
	int rb=0, i=0;
	char data[50];

	/* take away leading spaces or 10 or 13 or tabs or whatever */
	while (c==' ' || c=='\t' || c==10 || c==13) {
		if (-1 == (rb = recv (fd, &c, 1, 0))) return -1;
	}

	while (c!=' ' && c!='\t' && c!=10 && c!=13 && i < 49) {       
		data[i++] = c;
		if (-1 == (rb = recv (fd, &c, 1, 0))) goto done;
	}
done:
	data[i]=0;

	*nr = atof(data);
	return 1;
}

int get_data_head (int fd, double *tstart, int *size)
{
	double s;
	if (-1 == getdbl (fd, &s)) return -1;
	if (-1 == getdbl (fd, tstart)) return -1;
	*size = (int)s;

	return 0;
}

struct telejob *init_telejob (int fd, char strm)
{
	struct telejob *tj = (struct telejob *)malloc (sizeof(struct telejob));
	int i;
	if (!tj) return NULL;

	tj->fd = fd;
	tj->strm = strm;
	tj->ephem = NULL;

	if (-1 == get_data_head (fd, &(tj->tstart), &(tj->ephsize))) {
		free (tj);
		return NULL;
	}	

	printf ("mark\n"); fflush (stdout);

	if (tj->tstart == 0.0) tj->tstart = get_time();

	if (!strm) {
		/* no streaming -- read all data */
		tj->ephem = (eph_t*) malloc (sizeof(eph_t)*tj->ephsize);
		if (!tj->ephem) {
			free (tj);
			return NULL;
		}
		for (i = 0; i < tj->ephsize; i++) {
			printf ("reading position %d of %d\n",
				i+1, tj->ephsize);
			if (-1 == getdbl (tj->fd, &(tj->ephem[i].time)))
				goto fail;
			tj->ephem[i].time += tj->tstart;
			if (-1 == getdbl (tj->fd, &(tj->ephem[i].azm)))
				goto fail;
			if (-1 == getdbl (tj->fd, &(tj->ephem[i].elv)))
				goto fail;
			if (-1 == getdbl (tj->fd, &(tj->ephem[i].dst)))
				goto fail;			
		}
		goto ok;
		fail:
			free (tj->ephem);
			free (tj);
			return NULL;				
		ok:
			tj->ephpos = 0;
			printf ("batch read ok\n");
			return tj;
	} else {
		/* streaming -- don't read anyting */
		tj->fd = fd;
	}
	
	return tj;
}

void close_telejob (struct telejob *tj)
{
	if (tj->fd > -1) close (tj->fd);
	if (tj->ephem) free (tj->ephem);
	free (tj);
}

int get_eph (struct telejob *tj, eph_t *e)
{
	if (tj->strm) { /* streaming */
		if (-1 == getdbl (tj->fd, &e->time)) return -1;
		printf ("setting time to %f + %f\n", e->time, tj->tstart);
		e->time += tj->tstart;
		if (-1 == getdbl (tj->fd, &e->azm)) return -1;
		if (-1 == getdbl (tj->fd, &e->elv)) return -1;
		if (-1 == getdbl (tj->fd, &e->dst)) return -1;
		return 0;
	} else { /* no streaming */
		if (tj->ephpos >= tj->ephsize) return -1;
		else {
			if (!tj) return -1;
			*e = tj->ephem[tj->ephpos++];
		}
		return 0;
	}
	return -1;
}
