/* *************************************************************** */
/* Bibliothek zur MON03-Motorsteuerung							   */
/* TAB-Spacing = 4												   */
/* *************************************************************** */

/* *************************************************************** */
/* INCLUDE-Files												   */
/* *************************************************************** */

#include <dos.h>
#include <math.h>
#include <lib/oeslib.h>
#include "MON03H.HPP"

/* *************************************************************** */
/* Globale Variablen											   */
/* *************************************************************** */

/* Vorbesetzung der Register-Merkvariablen da                      */
/* die IO-Register nur schreibbar sind                             */
unsigned char	MotOVal = 0x00;
unsigned char	AuxVal  = 0x00;
unsigned char	DACVal  = 0x00;

/* *************************************************************** */
/* Es folgen NICHT hardwareabhaengige Routinen und Einstellungen   */
/* *************************************************************** */

// Ruecksetzen der Steuerung
// Die Steuerung wird per manuellem RESET komplett zurueckgesetzt
// alle Inhalte der Microcontroller gehen verloren!

void MON_Reset(void)
{
  RESET_low;    // Aktivieren der /RESET-Leitung der Motor-Microcontroller
  RESPIC_low;   // Aktivieren der /RESET-Leitung des Uhren-Microcontroller

  RW_1_low;		// Initialisieren der Interfaceleitungen
  RW_1_high;
  RW_2_low;
  RW_2_high;
  RW_3_low;
  RW_3_high;
  RW_4_low;
  RW_4_high;

  DI_1_low;
  DI_1_high;
  DI_2_low;
  DI_2_high;
  DI_3_low;
  DI_3_high;
  DI_4_low;
  DI_4_high;

  PSTR_low;         // Initialisieren der Schnittstelle zur ms-Uhr
  PSTR_high;
  PA0_low;
  PA0_high;
  PA1_low;
  PA1_high;


  RESPIC_high;		// Freigeben der Microcontroller
  RESET_high;

  SetDAC(0);		// Initialiseren des DAC

  double time; time = MON_GetClk();
}

/* *************************************************************** */

// DAC: Uo = 0..5 V, Ro = 500 Ohm
// Fuer Sprachausgabe an 4 pol. Stecker (Uo, GND, GND, GND) auf der
// 8-bit PC-Steckkarte

void MON_SetDAC(unsigned char value)
{
  SetDAC(value);
}

/* *************************************************************** */

// real-time will be returned in s (accuracy is 25.6us)
double MON_GetClk (void)
{
  double PIC_Count = 0.;

  for (int i=0; i<20; i++)
  {
	PA0_low;                 			// Adresse 0 zum PIC fuer erstes und
	if (i != 0) PA0_high;               // Adresse 1 fuer weitere Nibbles
	PA1_low;
	PSTR_low;							// Handshake durchfuehren, dass
//  printf("a:%d\n",i);
	while( (GetPicData&0x80)!=0 ) {};   // der PIC das MSNibble auch anlegt
	PSTR_high;
//  printf("b:%d\n",i);
	while( (GetPicData&0x80)==0 ) {};
	PIC_Count += (double)(GetPicData&0x0f) * pow( 16., double(19-i) );
  }

  return( PIC_Count / (2.e+7/512) );
}

/* *************************************************************** */

// Fuer Motor 1
// Speed:      spd in cycles/sec (max. 9764cycles/sec = 39056FullSteps/sec)
// PowerDown:  pd=0 ->50% pd=1 ->100% of current set by resistor
// Position:   *pos in cycles
// Limits:     *lim (bit0:upper switch,   bit1:lower switch)
//                  (value 1 for open and +5V, 0 for shortcut to GND)
// Attention:  time for one send- or receive-loop must be executed in
//             less than 12us !!!

int MON_TransCeive1(double spd, int pd, double *pos, char *lim)
{
  int 			i, val[50];

  val[0] = 0;                             // Dummy-bit
  val[1] = 0; if (pd!=0) { val[1]=1; };   // /PowerDown-bit

  spd   = spd/(2.e+7/512)*256;            // normalizing to one PIC-cycle
  if (spd >  63) spd= 63;
  if (spd < -63) spd=-63;
  long int buffs32 = (long int)(spd*pow(2., 24.)); // calc fix-point 1byte.3byte

  for (i=2; i<34; i++)
  {
	val[i] = 0; if((buffs32>>(33-i))&0x01) val[i]=1;
  }

  IntOff();							/* interrupts off   */

  while(GetCLK_1==0) { };  			/* set write to pic */
  RW_1_low;
  while(GetCLK_1!=0) { };

  // Attention:  one run through following loop must be executed in
  //             less than 12us !!!
  for (i=0; i<34; i++)				/* Send speed to PIC */
  {
	while(GetCLK_1==0) { };
	if (val[i]==0) { DI_1_low;  }
	else           { DI_1_high; }
	while(GetCLK_1!=0) { };
  }

  // Attention:  one run through following loop must be executed in
  //             less than 12us !!!
  for (i=0; i<1; i++)				/* Send dummy bits to PIC */
  {
	while(GetCLK_1==0) { };
	while(GetCLK_1!=0) { };
  }

  while(GetCLK_1==0) { };  			/* set read */
  RW_1_high;
  while(GetCLK_1!=0) { };

  while(GetCLK_1==0) { };
  while(GetCLK_1!=0) { };

  // Attention:  one run through following loop must be executed in
  //             less than 12us !!!
  buffs32 = 0;						/* receive pos from PIC */
  for (i=0; i<32; i++)
  {
	while(GetCLK_1==0) { };
	buffs32 = buffs32<<1;
	buffs32 |=  GetDO_1;
	while(GetCLK_1!=0) { };
  }

  // Attention:  one run through following loop must be executed in
  //             less than 12us !!!
  unsigned char buffu8 = 0; 		/* receive limits from PIC */
  for (i=0; i<8; i++)
  {
	while(GetCLK_1==0) { };
	buffu8 = buffu8<<1;
	buffu8 |=  GetDO_1;
	while(GetCLK_1!=0) { };
  }

  IntOn();						    /* interrupts on	*/

  *pos = (double)buffs32 / 256.;
  *lim = buffu8>>6;

  return(0);
}

/* *************************************************************** */

// Fuer Motor 2
// Speed:      spd in cycles/sec (max. 9764cycles/sec = 39056FullSteps/sec)
// PowerDown:  pd=0 ->50% pd=1 ->100% of current set by resistor
// Position:   *pos in cycles
// Limits:     *lim (bit0:upper switch,   bit1:lower switch)
//                  (value 1 for open and +5V, 0 for shortcut to GND)

int MON_TransCeive2(double spd, int pd, double *pos, char *lim)
{
  int 			i, val[50];

  val[0] = 0;                             // Dummy-bit
  val[1] = 0; if (pd!=0) { val[1]=1; };   // /PowerDown-bit

  spd   = spd/(2.e+7/512)*256;            // normalizing to one PIC-cycle
  if (spd >  63) spd= 63;
  if (spd < -63) spd=-63;
  long int buffs32 = (long int)(spd*pow(2., 24.)); // calc fix-point 1byte.3byte

  for (i=2; i<34; i++)
  {
	val[i] = 0; if((buffs32>>(33-i))&0x01) val[i]=1;
  }

  IntOff();

  while(GetCLK_2==0) { };  			/* set write to pic */
  RW_2_low;
  while(GetCLK_2!=0) { };

  for (i=0; i<34; i++)				/* Send speed to PIC */
  {
	while(GetCLK_2==0) { };
	if (val[i]==0) { DI_2_low;  }
	else           { DI_2_high; }
	while(GetCLK_2!=0) { };
  }

  for (i=0; i<1; i++)				/* Send dummy bits to PIC */
  {
	while(GetCLK_2==0) { };
	while(GetCLK_2!=0) { };
  }

  while(GetCLK_2==0) { };  			/* set read */
  RW_2_high;
  while(GetCLK_2!=0) { };

  while(GetCLK_2==0) { };
  while(GetCLK_2!=0) { };

  buffs32 = 0;						/* receive pos from PIC */
  for (i=0; i<32; i++)
  {
	while(GetCLK_2==0) { };
	buffs32 = buffs32<<1;
	buffs32 |=  GetDO_2;
	while(GetCLK_2!=0) { };
  }
  unsigned char buffu8 = 0; 		/* receive limits from PIC */
  for (i=0; i<8; i++)
  {
	while(GetCLK_2==0) { };
	buffu8 = buffu8<<1;
	buffu8 |=  GetDO_2;
	while(GetCLK_2!=0) { };
  }
  IntOn();

  *pos = (double)buffs32 / 256.;
  *lim = buffu8>>6;

  return(0);
}

/* *************************************************************** */

// Fuer Motor 3
// Speed:      spd in cycles/sec (max. 9764cycles/sec = 39056FullSteps/sec)
// PowerDown:  pd=0 ->50% pd=1 ->100% of current set by resistor
// Position:   *pos in cycles
// Limits:     *lim (bit0:upper switch,   bit1:lower switch)
//                  (value 1 for open and +5V, 0 for shortcut to GND)

int MON_TransCeive3(double spd, int pd, double *pos, char *lim)
{
  int 			i, val[50];

  val[0] = 0;                             // Dummy-bit
  val[1] = 0; if (pd!=0) { val[1]=1; };   // /PowerDown-bit

  spd   = spd/(2.e+7/512)*256;            // normalizing to one PIC-cycle
  if (spd >  63) spd= 63;
  if (spd < -63) spd=-63;
  long int buffs32 = (long int)(spd*pow(2., 24.)); // calc fix-point 1byte.3byte

  for (i=2; i<34; i++)
  {
	val[i] = 0; if((buffs32>>(33-i))&0x01) val[i]=1;
  }

  IntOff();

  while(GetCLK_3==0) { };  			/* set write to pic */
  RW_3_low;
  while(GetCLK_3!=0) { };

  for (i=0; i<34; i++)				/* Send speed to PIC */
  {
	while(GetCLK_3==0) { };
	if (val[i]==0) { DI_3_low;  }
	else           { DI_3_high; }
	while(GetCLK_3!=0) { };
  }

  for (i=0; i<1; i++)				/* Send dummy bits to PIC */
  {
	while(GetCLK_3==0) { };
	while(GetCLK_3!=0) { };
  }

  while(GetCLK_3==0) { };  			/* set read */
  RW_3_high;
  while(GetCLK_3!=0) { };

  while(GetCLK_3==0) { };
  while(GetCLK_3!=0) { };

  buffs32 = 0;						/* receive pos from PIC */
  for (i=0; i<32; i++)
  {
	while(GetCLK_3==0) { };
	buffs32 = buffs32<<1;
	buffs32 |=  GetDO_3;
	while(GetCLK_3!=0) { };
  }
  unsigned char buffu8 = 0; 		/* receive limits from PIC */
  for (i=0; i<8; i++)
  {
	while(GetCLK_3==0) { };
	buffu8 = buffu8<<1;
	buffu8 |=  GetDO_3;
	while(GetCLK_3!=0) { };
  }
  IntOn();

  *pos = (double)buffs32 / 256.;
  *lim = buffu8>>6;

  return(0);
}

/* *************************************************************** */

// Fuer Motor 4
// Speed:      spd in cycles/sec (max. 9764cycles/sec = 39056FullSteps/sec)
// PowerDown:  pd=0 ->50% pd=1 ->100% of current set by resistor
// Position:   *pos in cycles
// Limits:     *lim (bit0:upper switch,   bit1:lower switch)
//                  (value 1 for open and +5V, 0 for shortcut to GND)

int MON_TransCeive4(double spd, int pd, double *pos, char *lim)
{
  int 			i, val[50];

  val[0] = 0;                             // Dummy-bit
  val[1] = 0; if (pd!=0) { val[1]=1; };   // /PowerDown-bit

  spd   = spd/(2.e+7/512)*256;            // normalizing to one PIC-cycle
  if (spd >  63) spd= 63;
  if (spd < -63) spd=-63;
  long int buffs32 = (long int)(spd*pow(2., 24.)); // calc fix-point 1byte.3byte

  for (i=2; i<34; i++)
  {
	val[i] = 0; if((buffs32>>(33-i))&0x01) val[i]=1;
  }

  IntOff();

  while(GetCLK_4==0) { };  			/* set write to pic */
  RW_4_low;
  while(GetCLK_4!=0) { };

  for (i=0; i<34; i++)				/* Send speed to PIC */
  {
	while(GetCLK_4==0) { };
	if (val[i]==0) { DI_4_low;  }
	else           { DI_4_high; }
	while(GetCLK_4!=0) { };
  }

  for (i=0; i<1; i++)				/* Send dummy bits to PIC */
  {
	while(GetCLK_4==0) { };
	while(GetCLK_4!=0) { };
  }

  while(GetCLK_4==0) { };  			/* set read */
  RW_4_high;
  while(GetCLK_4!=0) { };

  while(GetCLK_4==0) { };
  while(GetCLK_4!=0) { };

  buffs32 = 0;						/* receive pos from PIC */
  for (i=0; i<32; i++)
  {
	while(GetCLK_4==0) { };
	buffs32 = buffs32<<1;
	buffs32 |=  GetDO_4;
	while(GetCLK_4!=0) { };
  }
  unsigned char buffu8 = 0; 		/* receive limits from PIC */
  for (i=0; i<8; i++)
  {
	while(GetCLK_4==0) { };
	buffu8 = buffu8<<1;
	buffu8 |=  GetDO_4;
	while(GetCLK_4!=0) { };
  }
  IntOn();

  *pos = (double)buffs32 / 256.;
  *lim = buffu8>>6;

  return(0);
}

