#ifndef __STEPENG_H
#define __STEPENG_H

#include <linux/types.h>
#include <linux/stepeng_ioctl.h>

/*
 * Copyright (C) 2000 by Florin Boariu <florin@bnv-bamberg.de>
 *
 * Some basic information about the driver:
 *
 * It controls a O.E.S. PIC Controler for up to 4 high-precision
 * engines (a.k.a. step-engines). These engines can be digitally
 * controlled to have a certain speed. The granulation of the speed
 * depends on the number of steps an engine needs for a turn
 * (the higher the number the finer-grained the speed)
 * The speed can vary from very few steps/day (or maybe even week) to
 * several rotations (more than 30k degrees) per second.
 *
 * For abstraction reasons, this driver works with speed stored as
 * deg/sec in a double-precision floating point number (a.k.a. doubles :)
 *
 * The PIC controller receives cycles/25.6us in fixed-point numbers.
 * The allignment is 1byte.3byte.
 *
 * Further below are some macros to convert these values from one to
 * another. If you need to update this driver for another engine type,
 * all you have to do is change the constant values the multiplication
 * is done with.
 *
 * The driver does the conversion only once: when it sends the data
 * to the controler. Every other transformation inside this driver is
 * performed on the double-values.
 *
 * Watch the bit-sending and bit-receiving routines (there are two inline
 * functions). They _must_ be performed by the machine within 12us _at_ _most_.
 * That's a restriction of the PIC controler.
 *
 * Each of the engines has two limit-detectors attached. They cause
 * certain bits to be set when the drive attached to the engine hits
 * some edges. Beware that they have to be checked on periodically,
 * and in case of them being set, the engine has to be stopped!
 * 
 * The hardware-communication macros are based on Dr Frank Fleischmann's
 * <ff@fonline.de> reference software for MS-DOS.
 *
 * Oh, the hardware kinda sucks. We'll have to save the settings we have
 * written (the speed) aswell as the time, and compute the position of
 * the device dynamically. We cannot ask the device's status at random,
 * but only right after we have written.
 *
 * I still have no idea how to check the limits and make an emergency stop
 * when they occur.
 */

#define HPE_MAX_DEV 4

#define HPE_VERSION "0.1.100" /* pre-version to 0.2.0 */

/*
 * Version history:
 * up to 0.1.0: internal/not working/development
 * 0.1.0:       the infamous Nuernberg-Version (march 6th, 2001)
 * 0.2.0:       the new driver, 2.4-compatible.
 */

#define HPE_MOV_INCR +1
#define HPE_MOV_NONE  0
#define HPE_MOV_DECR -1

struct hpe_devinfo {
	int    open;                                   /* is the device open? */
	struct hpe_accel_ramp accel_ramp;              /* the acceleration ramp */
	int ramp_pos;                                  /* the position in the ramp */
	signed char last_move;                         /* whether the last move was incr, decr or nothing */
	struct hpe_engine_state estate;                /* the engine state */
	double spd_fac, pos_fac;                       /* speed factor and position factor */
};

/*
 * struct to deal with the position information of the engine
 *
 *  Engine:
 *  xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx . xxxxxxxx xxxxxxxx xxxxxxxx
 *  |<------------ LEFT ------------->|   |<-------- RITE -------->|
 *  
 *  Struct:
 *  xxxxxxxx xxxxxxxx xxxxxxxx xxxxxxxx . oooooooo xxxxxxxx xxxxxxxx xxxxxxxx
 *  |<------------ LEFT ------------->|   <-SUPL-> |<-------- RITE -------->|
 */
struct hpe_engine_pos {
	uint32_t left;   /* whole steps */
	uint32_t rite;   /* fractions */
};

struct hpe_time {
	uint32_t sec;
	uint32_t usec;
};

/* computes the difference in msecs (??) */
#define HPE_MSDIFF(t1,t2) (t1>t2? (0xffffffff-t1+t2)*10/390624 : (t2-t1)*10/3906264)

/* moves the SUPL-part of an engine_pos to the LEFT (normalize) */
void hpe_normpos (struct hpe_engine_pos *pos);

#define SUCCESS 0

/* the device file name */
#define HPE_NAME "hpe"

/* map the position from 0..ffffffff to -80000000..7fffffff */
#define POSITION_MAP(pos) (pos-0x80000000)

/* extern struct hpe_accel_ramp accel_ramp; */

/*
 * 360 degrees = 400 steps (200 steps on some engines)
 *   1 cycle   =   4 steps
 * 256 msteps  =   1 step
 * 256 mcycles =   1 cycle (of course)
 *
 * the device receives velocity data in fixed point 1.3 bit
 * the velocity is specified in msteps/(20MHz/512) = msteps/25.6us
 *
 *     cycles           time         normalize to fixed point
 *
 * (deg/360*400/4) / (20MHz/512)   *      (256*2^24);  const: 30541.9896605
 *
 * Ahm, I changed the 'deg' to 'rad'. So a full turn will be 2*PI, not 360.
 * => new const:  1749927.10545
 *     nope, it's 874963.552725 (we do indeed have a 200 step engine)
 */

#ifndef PI
#define PI 3.14159265358979
#endif

#define HPE_SPD_DEV(e,rad) ((uint32_t)(rad*e.spd_fac))  /* radients to device - speed */
#define HPE_SPD_RAD(e,dev) ((double)   dev/e.spd_fac)   /* device to radients - speed */

#define HPE_POS_DEV(e,rad) ((uint32_t)(rad*e.pos_fac))  /* radients to device - position */
#define HPE_POS_RAD(e,dev) ((double)   dev/e.pos_fac)   /* device to radients - position */

/*
 * hardware communication stuff
 */

#ifdef __BUILDING_HPE__

extern unsigned int HPEPort;
extern unsigned char HPEconIO;  /* contents of the i/o register */
extern unsigned char HPEconAux; /* contents of the aux register */
extern unsigned char HPEconDac; /* contents of the dac register */

/* the register */
#define HPEOut (HPEPort+0x00)
#define HPEIn  (HPEPort+0x00)
#define HPEAux (HPEPort+0x01)
#define HPEPic (HPEPort+0x01)
#define HPEDac (HPEPort+0x02)

/*
 * card I/O
 */
#define Out(val,reg) outb_p((val),(reg))
#define In(reg)      inb_p((reg))

#define HPErwLo(nr)     (HPEconIO & ~(0x01<<(nr<<1)))
#define HPErwHi(nr)     (HPEconIO |  (0x01<<(nr<<1)))
#define SetRW(nr,bit)   HPEconIO = (bit ? HPErwHi(nr) : HPErwLo(nr) ); Out (~HPEconIO,HPEOut)

#define HPEdiLo(nr)     (HPEconIO & ~(0x02<<(nr<<1)))
#define HPEdiHi(nr)     (HPEconIO |  (0x02<<(nr<<1)))
#define SetDI(nr,bit)   HPEconIO = (bit ? HPEdiHi(nr) : HPEdiLo(nr) ); Out (~HPEconIO,HPEOut)

#define AuxLo(bb)       (HPEconAux & ~bb)
#define AuxHi(bb)       (HPEconAux |  bb)
#define SetAux(bit,bb)  HPEconAux = (bit ? AuxHi(bb) : AuxLo(bb)); Out (HPEconAux,HPEAux)

#define SetPSTR(bit)    SetAux(bit,0x01)
#define SetPA0(bit)     SetAux(bit,0x02)
#define SetPA1(bit)     SetAux(bit,0x04)
#define SetRESPIC(bit)  SetAux(bit,0x40)

#define ResetHi          (HPEconAux & ~0x80)
#define ResetLo          (HPEconAux |  0x80)
#define SetReset(bit)    HPEconAux = (bit ? ResetHi : ResetLo); Out (HPEconAux,HPEAux)

#define SetDAC(v)        HPEconDac=v; Out(HPEconDac,HPEDac)

#define GetClk(nr)       ( (~In(HPEIn) >>  (nr<<1))    & 0x01)
#define GetOut(nr)       ( (~In(HPEIn) >> ((nr<<1)+1)) & 0x01)
#define GetPIC           (In(HPEPic))

/* kernel module functions */
int hpe_init (void);

/* controler functions */
void    hpe_reset_pic   (void);
double  hpe_get_clock   (void);
inline void    hpe_put_dac     (unsigned char c);

/* enigne functions */
inline void    hpe_send_bit    (uint8_t engine_nr, uint8_t bit);
inline uint8_t hpe_recv_bit    (uint8_t engine_nr);
inline void hpe_engine_send (uint8_t engine_nr, double speed, int power_down);
inline void hpe_engine_recv (uint8_t engine_nr, uint32_t *steps, uint8_t *lim);
inline void hpe_set_speed   (uint8_t engine_nr, double gps);
double hpe_goto_speed  (uint8_t engine_nr, double gps);
void hpe_wait        (double sec);
void hpe_ctrl_engine (uint8_t nr, double speed,
		      struct hpe_engine_state *stat);

/* convenience functions */
void hpe_mkramp (struct hpe_accel_ramp *ramp,
		 int size, double max_t, double max_sp);

#endif /* __BUILDING_HPE__ */

#endif /* __STEPENG_H */

