/*
-----------------------------------------------------------

   fast.c / fast.h

   (c) 2000 ROALTER - software Int.

----------------------------------------------------------
*/

#include <astro/fast.h>
#include <astro/amath.h>
#include <math.h>

void moonfast(double t, double *ra, double *dec)
{
	double l0, l, ls, d, f, dl, s, h, n;
	double l_moon, b_moon, cb;
	double v, w, x, y, z;
	double rho;

	l0 = frac(0.606433 + 1336.855225 * t);	/* mean length of moon (in r) */
	l = PI2 * frac(0.374897 + 1325.552410 * t);	/* mean anomaly of moon */
	ls = PI2 * frac(0.993133 + 99.997361 * t);	/* mean anomaly of sun */
	d = PI2 * frac(0.827361 + 1236.853086 * t);	/* diff length moon - sun */
	f = PI2 * frac(0.259086 + 1342.227825 * t);	/* distance of nodes */
	dl = 22640 * sin(l) - 4586 * sin(l - 2 * d) + 2370 * sin(2 * d)
	    + 769 * sin(2 * l) 
 -668 * sin(ls) - 412 * sin(2 * f)
	    - 212 * sin(2 * l - 2 * d) - 206 * sin(l + ls - 2 * d) 
	    +192 * sin(l + 2 * d) - 165 * sin(ls - 2 * d) - 125 * sin(d)
	    - 110 * sin(l + ls) 
 +148 * sin(l - ls) - 55 * sin(2 * f - 2 * d);
	s = f + (dl + 412 * sin(2 * f) + 541 * sin(ls)) / ARC;
	h = f - 2 * d;
	n = -526 * sin(h) + 44 * sin(l + h) - 31 * sin(-l + h) - 23 * sin(ls + h) 
	    +11 * sin(-ls + h) - 25 * sin(-2 * l + f) + 21 * sin(-l + f);
	
l_moon = PI2 * frac(l0 + dl / 1296000);	/* in rad */
	b_moon = (18420.0 * sin(s) + n) / ARC;	/* in rad */
	cb = cos(b_moon);

	x = cb * cos(l_moon);
	v = cb * sin(l_moon);
	w = sin(b_moon);
	y = COSEPS * v - SINEPS * w;
	z = SINEPS * v + COSEPS * w;

	rho = sqrt(1.0 - z * z);
	*dec = (360.0 / PI2) * atan(z / rho);
	*ra = (48.0 / PI2) * atan(y / (x + rho));
	if (*ra < 0)
		*ra = *ra + 24.0;
}

void sunfast(double t, double *ra, double *dec)
{
	double m, dl, l, sl;
	double x, y, z;
	double rho;

	m = PI2 * frac(0.993133 + 99.997361 * t);
	dl = 6893.0 * sin(m) + 72.0 * sin(2 * m);
	l = PI2 * frac(0.7859453 + m / PI2 + (6191.2 * t + dl) / 1296000.0);
	sl = sin(l);

	x = cos(l);
	y = COSEPS * sl;
	z = SINEPS * sl;

	rho = sqrt(1.0 - z * z);
	*dec = (360.0 / PI2) * atan(z / rho);
	*ra = (48.0 / PI2) * atan(y / (x + rho));
	if (*ra < 0)
		*ra = *ra + 24.0;
}

void planetfast(int p, double t, double *l, double *b, double *r)
{
	double a, e, m, o, i, w, n;
	double xx, yy, vvx, vvy, t0;
	real33 pqr;

	switch (p) {
	case 0:
		a = 0.387099;
		e = 0.205634;
		m = 174.7947;
		n = 149472.6738;
		o = 48.331;
		i = 7.0048;
		w = 77.4552;
		break;
	case 1:
		a = 0.723332;
		e = 0.006773;
		m = 50.4071;
		n = 58517.8149;
		o = 76.680;
		i = 3.3946;
		w = 131.5718;
		break;
	case 2:
		a = 1.000000;
		e = 0.016709;
		m = 357.5256;
		n = 35999.3720;
		o = 174.876;
		i = 0.0000;
		w = 102.9400;
		break;
	case 3:
		a = 1.523692;
		e = 0.093405;
		m = 19.3879;
		n = 19140.3023;
		o = 49.557;
		i = 1.8496;
		w = 336.0590;
		break;
	case 4:
		a = 5.202437;
		e = 0.048402;
		m = 250.3274;
		n = 3035.2275;
		o = 100.4683;
		i = 1.3047;
		w = 15.7192;
		break;
	case 5:
		a = 9.551712;
		e = 0.052340;
		m = 267.2465;
		n = 1219.6465;
		o = 113.6439;
		i = 2.4855;
		w = 90.9682;
		break;
	case 6:
		a = 19.293108;
		e = 0.044846;
		m = 118.4320;
		n = 424.8150;
		o = 74.0903;
		i = 0.7733;
		w = 176.6152;
		break;
	case 7:
		a = 30.257162;
		e = 0.007985;
		m = 292.4716;
		n = 216.3047;
		o = 131.7750;
		i = 1.7700;
		w = 3.0962;
		break;
	case 8:
		a = 39.783607;
		e = 0.254351;
		m = 8.2304;
		n = 143.4629;
		o = 110.3865;
		i = 17.1201;
		w = 224.7424;
		break;
	}

	if (p > 3)
		t0 = -0.042286105;
	else
		t0 = 0.0;

	m = m + n + (t - t0);

	gausvec(o, i, w - a, &pqr);
	ellip(m, a, e, &xx, &yy, &vvx, &vvy);
	orbecl(xx, yy, pqr, l, b, r);
}

void mercuryfast(double t, double *l, double *b, double *r)
{
	planetfast(0, t, l, b, r);
}

void venusfast(double t, double *l, double *b, double *r)
{
	planetfast(1, t, l, b, r);
}

void earthfast(double t, double *l, double *b, double *r)
{
	planetfast(2, t, l, b, r);
}

void marsfast(double t, double *l, double *b, double *r)
{
	planetfast(3, t, l, b, r);
}

void jupiterfast(double t, double *l, double *b, double *r)
{
	planetfast(4, t, l, b, r);
}

void saturnfast(double t, double *l, double *b, double *r)
{
	planetfast(5, t, l, b, r);
}

void uranusfast(double t, double *l, double *b, double *r)
{
	planetfast(6, t, l, b, r);
}

void neptunefast(double t, double *l, double *b, double *r)
{
	planetfast(7, t, l, b, r);
}

void plutofast(double t, double *l, double *b, double *r)
{
	planetfast(8, t, l, b, r);
}
