#include <sky/core/joblistdir.hh>
#include <errno.h>
#include <stdexcept>
#include <fstream>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/file.h>
#include <stream.h>
#include <cfg/cfg.hh>

#define JOBNAME "%1.3f+%1.3f:%s"

JobListDir::JobListDir()
	:dir(NULL)
{
}

JobListDir::JobListDir(const string &d)
	:dir(NULL)
{
	this->init (d);
}

JobListDir::~JobListDir ()
{
	this->cleanup();
}

void JobListDir::init (const string &i)
{
	map<string,string> cfg;
	strToMap (i, cfg);
	this->init (cfg);
}

void JobListDir::init (map<string,string> &cfg)
{
	if (dir) cleanup();
	dir = opendir(cfg["dirname"].c_str());
	if (!dir) {
		cerr << cfg["dirname"] << ": " << strerror(errno) << endl;
		throw runtime_error("JobListDir::init: Cannot open directory.");
	}
	
	dirname = cfg["dirname"];
	dirname += '/';
}

void JobListDir::cleanup()
{
	closedir(dir);
}

bool JobListDir::closestTo (Time ref, bool after, JobData &data)
{
	struct dirent *d;
	Time found=-1.0;

	rewinddir(dir);

 loop:
	while (d = readdir(dir)) {

		Time st=0.0, le=0.0;
		string mod;
		if (!parseName (d->d_name, st, le, mod)) {
			//cerr << "could not parse " << d->d_name << endl;
			goto loop;
		}

		//printf ("checking: %s (%f + %f <=> %f and %f)\n", d->d_name, st, le, ref, found);

		if ((after && st>=ref && (found<0.0 || st<found))  ||
		    (!after && st<ref && (found<0.0 || st>found))) {
			// we found it
			data.setStart(st);
			data.setLen(le);
			data.setMod(mod);
			data.setId(d->d_name);
			found = st;
		}
	}
	return !(found<0.0);
	//return found>0.0;
}

bool JobListDir::parseName (const char *name, Time &st, Time &le, string &mod)
{
	// parse the file name: ss.ss+ll.ll:module
	const char *str = name;

	//printf ("parsing: %s\n", str);
	
	st = atof(str);
	if (st == 0.0) return false;
	while (*str && *str++ != '+'){};
	if (!*str) return false;
	//printf ("le (str): %s\n", str);
	le = atof(str);
	while (*str && *str++ != ':'){};
	if (!*str) return false;
	mod = str;
	
	return true;
}

void JobListDir::validate(JobData &data)
{
	char name[100];
	snprintf (name, 99, JOBNAME,
		  data.start(), data.len(), data.module().c_str());
	//cerr << "reading from: " << dirname + name << endl;
	ifstream f ((dirname+name+"/options").c_str());
	dataIn (f, data);
}

bool JobListDir::empty(void)
{
	struct dirent *d;
	rewinddir(dir);
	while (d = readdir(dir)) {
		if (isdigit(d->d_name[0])) { return false; }
	}
	return true;
}

bool JobListDir::earliest(JobData &data)
{
	bool ret = closestTo (0.0, true, data);
	if (ret) cout << "JobListDir: earliest found" << endl;
	return ret;
}

void JobListDir::discardEarliest(void)
{
	JobData d;
	this->closestTo (0.0, true, d);
	char name[100];
	snprintf (name, 99, JOBNAME,
		  d.start(), d.len(), d.module().c_str());
	unlink ((dirname+name+"/options").c_str());
	unlink ((dirname+name+"/output").c_str());
	rmdir ((dirname+name).c_str());
}

bool JobListDir::schedule(const JobData &data)
{
	JobData iprev, iafter;
	bool bprev = this->closestTo(data.start(), false, iprev);
	bool bafter = this->closestTo(data.start(), true, iafter);
	bool ok=true;

	//printf ("%d: %f >= %f\n", bprev, iprev.start()+iprev.len(), data.start());
	//printf ("%d: %f >= %f\n", bafter, data.start()+data.len(), iafter.start());

	// check for place
	if (bprev && iprev.start()+iprev.len() >= data.start()) ok=false;
	if (bafter && data.start()+data.len() >= iafter.start()) ok=false;

	if (ok) {
		//
		// each job will be saved in a directory.
		// the directory will have at most the following files:
		//
		// options: all the job options (time, object...)
		// output:  the result (a.k.a. the image)
		//
		char name[100];
		snprintf (name, 99, JOBNAME,
			  data.start(), data.len(), data.module().c_str());
		if (-1 == mkdir ((dirname+name).c_str(), 0775)) {
			perror ("mkdir");
			throw runtime_error ("JobListDir::schedule: could not create job directory");
		}
		ofstream f ((dirname+name+"/options").c_str());
		//data["output"] = dirname+name+"/output"; // save the output file name for later
		if (f.fail()) throw runtime_error ("JobListDir::schedule: could not create job options file");
		//if (f.fail()) {
		//	cerr << "error creating file" << endl; cerr.flush();
		//}
		//cout << (dirname+name) << endl; cout.flush();
		this->dataOut (f, data);
		f.close();
		return true;
	} else return false;
}

void JobListDir::lock(void)
{
	flock(dirfd(dir), LOCK_EX);
}

void JobListDir::unlock(void)
{
	flock(dirfd(dir), LOCK_UN);
}

void JobListDir::dataIn (istream &in, JobData &data)
{
	data.clear();
	char line[256];
	string left, right;
	bool r=false;
	
	while (!(in.getline(line,255)).eof()) {
		//cerr << "line: " << line << endl;
		if (*line && *line!='#') {
			if (r) {
				right+=line;
				line[0]=0;
			} else {
				r=true;
				left=line;
				right.erase();
				line[0]=0;
			}
		} else if (!*line) {
			if (left.length() && right.length()) {
				data[left] = right;
				left.erase();
			}
			r=false;
		}
	}
	if (left.length()) {
		if (line[0]) right += line;
		data[left] = right;
		left.erase();
	}
}
	
void JobListDir::dataOut (ostream &out, const JobData &data)
{
	for (JobData::const_iterator i = data.begin();
	     i != data.end(); i++) {
		out << (*i).first << endl << (*i).second << endl << endl;
	}
	out << endl;
}

JobListDir *factory_JobListDir (void)
{
	return new JobListDir;
}

bool JobListDir::firstItem (JobData &data)
{
	struct dirent *d;
	rewinddir(dir);
	do {
		d = readdir(dir);
		if (!d) return false;
	} while (!isdigit(d->d_name[0]));

	Time st, le;
	string mod;
	
	parseName (d->d_name, st, le, mod);
	data.setStart(st);
	data.setLen(le);
	data.setMod(mod);
	data.setId(d->d_name);

	return true;
}

bool JobListDir::nextItem (JobData &data)
{
	struct dirent *d = readdir (dir);

	if (!d) return false;

	Time st, le;
	string mod;
      
	parseName (d->d_name, st, le, mod);
	data.setStart(st);
	data.setLen(le);
	data.setMod(mod);
	data.setId(d->d_name);
	
	return true;
}

bool JobListDir::find (const string &id, JobData &data)
{
	struct dirent *d;
	rewinddir(dir);
 loop:
	while (d = readdir(dir)) {
		//cerr << "comparing: " << id << " and " << d->d_name << endl;
		if (strcmp(id.c_str(), d->d_name)) goto loop;

		//cerr << "they match" << endl;

		Time st=0.0, le=0.0;
		string mod;

		if (!parseName (d->d_name, st, le, mod)) return false;

		data.setStart(st);
		data.setLen(le);
		data.setMod(mod);
		data.setId(d->d_name);
		return true;
	}
	return false;
}

bool JobListDir::discard (const string &id)
{
	//cerr << "removing: " << dirname+id << endl;
	unlink ((dirname+id+"/options").c_str());
	unlink ((dirname+id+"/output").c_str());
	rmdir ((dirname+id).c_str());
}
