#include <sky/drv/joblistdrv.hh>
#include <errno.h>
#include <stdexcept>
#include <fstream>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stream.h>

#include <cfg/cfg.hh>

JobListDrv::JobListDrv()
	:memory(NULL),ring(NULL),info(NULL),semkey(0),semid(0),tmp("/tmp/")
{
}

JobListDrv::JobListDrv(const string &d)
	:memory(NULL),ring(NULL),info(NULL),semkey(0),semid(0),tmp("/tmp/")
{
	this->init (d);
}

JobListDrv::~JobListDrv ()
{	
	this->cleanup();
}

void JobListDrv::init (const string &c)
{	
	map<string,string> cfg;
	strToMap (c, cfg);
	this->init (cfg);
}

/*
 * Parameters:
 * timebase:  time base in epoch time. every time value will be
 *            added to the time base in order to have the real start time
 *            "now" assumes now. really.
 * jobs:      size of the ring buffer in elements (maximum jobs)
 * fast:      when set to "true" or "yes", fast insertion is done.
 *            this is the only implementation supported ATM. still, it
 *            has to be explicitely specified in order to make the sender
 *            application aware of this situation.
 * semkey:    the key of the semaphore set to use.
 * shfile:    file to be mmaped and used as shared mem. "any" will have
 *            a temporary file name used instead.
 * newshf:    "yes" or "true" will cause the existing shared file (if one
 *            indeed exists) to be overwritten
 * cleanup:   cleanup on exit: if not true or yes, shared file will not be
 *            erased on exit (=> session can be resored later)
 */
void JobListDrv::init (map<string,string> &cfg)
{	
	// first thing to do:
	// get the semaphore and get a lock.
	// _then_ start init procedure (read file etc)
	// _then_ unlock. normally we're not responsible for locking,
	// but we do it here since the user has no possibility of locking.
	// another idea would be to move the lock completely
	// inside this class...

	if (!(cfg["shfile"].length())) {
		cerr << "JobListDrv::init: shfile not specified" << endl;
		throw runtime_error ("JobListDrv::init: shfile not specified");
	}
// if (-1 == (semkey = ftok (cfg["shfile"].c_str(), 'a'))) {
// 		cerr << "JobListDrv::init: ftok (" << cfg["shfile"] << "): "
// 		     << strerror (errno) << endl;
// 		throw runtime_error ("JobListDrv::init: cannot get sem key");
//	}
	
	semkey = stoi (cfg["semkey"]);
	cout << "semkey: " << semkey << endl; cout.flush();

	if (-1 == (semid = semget (semkey, 1, IPC_CREAT | 0660))) {
		cerr << "JobListDrv::init: semget: " << strerror (errno) << endl; cerr.flush();
		throw runtime_error ("JobListDrv::init: cannot create semaphore");
	}
	cout << "semid: " << semid << endl; cout.flush();

	this->lock();

	shfname = cfg["shfile"];
	if (!shfname.length()) shfname = "shfile";
	int msize =
		sizeof(struct list_info)+
		sizeof(struct data_ring)+
		sizeof(struct job)*stoi(cfg["jobs"]);	

	// try to open it first, w/o creating it. if the file arleady exists,
	// then some session is already running.
	if ((-1 == (shfd = open (shfname.c_str(), O_RDWR))) ||
	    (cfg["newshf"] == "yes" || cfg["newshf"] == "true")) {       
		if (-1 == (shfd = open (shfname.c_str(), O_RDWR | O_CREAT))) {
			perror ("open");
			throw runtime_error ("JobListDir::init: cannot create shared file");
		}
		fchmod (shfd, 0644);
		// we have to fill that file to the needed amount,
		// or else a "bus error" will occur.
		for (int i=0; i < msize; i+= 10) 
			write (shfd, "          ", 10);

		create_shfile (shfd, stof(cfg["timebase"]), stoi(cfg["jobs"]));
		ownshf = true && (cfg["cleanup"]!="yes" && cfg["cleanup"]!="true");
	} else {
		read_shfile (shfd);
		ownshf = false;
	}

	fast = (cfg["fast"] == "yes" || cfg["fast"] == "true");

	this->unlock();
}

/*
 * generate necessary header information for the shfile (-> set up the file)
 */
bool JobListDrv::create_shfile (int fd,
				double tb,
				int jobs)
{
	int msize =
		sizeof(struct list_info)+
		sizeof(struct data_ring)+
		sizeof(struct job)*jobs;
	
	cerr << "buffer size: " << msize << endl; cerr.flush();

	if (MAP_FAILED == (memory = (unsigned char*)mmap (NULL, msize, PROT_READ | PROT_WRITE,
				    MAP_SHARED, fd, 0))) {
		cerr << "Unable to mmap " << (getShfDir()+shfname) << ": "
		     << strerror (errno) << endl;
		throw runtime_error ("JobListDrv::init: cannot mmap ring buffer file");
	}

	cerr << "initializing shared buffer: " << (void*)memory << endl; cerr.flush();

	info = (struct list_info*)memory;
	ring = (struct data_ring*)&(memory[sizeof(struct list_info)]);
	info->timebase = tb;
	info->memsize = msize;
	ring->size = jobs;
	ring->pos = 0;
	ring->end = 0;
	ring->notempty = false;
	ring->buf = (struct job*)&memory[sizeof(struct list_info)+sizeof(struct data_ring)];
	
	cerr << "create_shfile done" << endl; cerr.flush();
}

/*
 * read the header information from the shfile
 * and use it to set up 'this' for work with that file.
 */
bool JobListDrv::read_shfile (int fd)
{
	struct data_ring dr;
	struct list_info li;

	lseek (fd, 0, SEEK_SET);
	if (sizeof(struct list_info) <
	    read (fd, &li, sizeof (struct list_info))) {
		throw runtime_error ("JobListDrv::read_shfile: "
				     "cannot read list info from shared file");
	}
	if (sizeof(struct data_ring) <
	    read (fd, &dr, sizeof (struct data_ring))) {
		throw runtime_error ("JobListDrv::read_shfile: "
				     "cannot read data ring header from shared file");
	}

	if (MAP_FAILED == (memory = (unsigned char*)mmap (NULL, li.memsize, PROT_READ | PROT_WRITE,
				    MAP_SHARED, fd, 0))) {
		cerr << "Unable to mmap " << (getShfDir()+shfname) << ": "
		     << strerror (errno) << endl;
		throw runtime_error ("JobListDrv::init: cannot mmap ring buffer file");
	}
	cerr << "shfile mmapped." << endl;
	cerr.flush();
	
	info = (struct list_info*)memory;
	ring = (struct data_ring*)&(memory[sizeof(struct list_info)]);
	cerr << "shfile layout: " << info->memsize << " bytes for up to " << ring->size
	     << " jobs (from " << ring->pos << " to " << ring->end << " already set)" << endl;
	cerr.flush();
}

void JobListDrv::cleanup()
{
	this->lock();
	munmap (memory, info->memsize);
	close (shfd);
	if (ownshf) {
		cerr << "removing shared file" << endl;
		unlink ((getShfDir()+shfname).c_str());
	} else {
		cerr << "NOT removing shared file" << endl;
	}
	ring = NULL;
	info = NULL;
	memory = NULL;
	this->unlock();
	semctl (semid, 0, IPC_RMID, NULL);
}

void JobListDrv::validate(JobData &data)
{
	// not needed
}

bool JobListDrv::empty(void)
{
	return (ring->end==ring->pos) && (!ring->notempty);
}

bool JobListDrv::earliest(JobData &data)
{
	if (this->empty()) {
		return false;
	} else {
		dataCls (data, &(ring->buf[ring->pos]));
		return true;
	}
}

void JobListDrv::discardEarliest(void)
{
	// i'm still trying to find out whether i'm incredibly stupid
	// or incredibly smart. mail me if you can tell...

	if ((ring->pos != ring->end) || ((ring->pos == ring->end) && ring->notempty)) {
		ring->notempty = false; // this doesn't mean ring really is empty!
		ring->pos = (ring->pos+1)%ring->size;
	}
	sleep (1);
}

/*
 * copy struct job to JobData.
 * we cannot store JobData in the file, so we have to store
 * struct jobs in there. but we use JobData everywhere in the
 * software...
 */
void JobListDrv::dataCls (JobData &c, const struct job *j)
{
	string mod ("skydrv");
	c.clear();
	c.setStart (j->start);
	c.setLen (j->len);
	c.setMod (mod);
	c["time"] = j->time;
	c["x"] = j->x;
	c["y"] = j->y;
	c["z"] = j->z;
}

/*
 * copy JobData to struct job
 */
void JobListDrv::dataStr (struct job *j, const JobData &c)
{
	j->start = c.start();
	j->len = c.len();
	strncpy (j->time, c.fetch("time").c_str(), 19);
	j->time[MAX_NR-1]=0;
	strncpy (j->x, c.fetch("x").c_str(), MAX_NR-1);
	j->x[MAX_NR-1]=0;
	strncpy (j->y, c.fetch("y").c_str(), MAX_NR-1);
	j->y[MAX_NR-1]=0;
	strncpy (j->z, c.fetch("z").c_str(), MAX_NR-1);
	j->z[MAX_NR]=0;
}

bool JobListDrv::schedule(const JobData &data)
{
	// WATCH IT! Have to check the time, and schedule it at the last position!
	// if fast==true, we will NOT do the checking, but rather assume that
	// the data arrives in the correct order (=> we simply push new values at the
	// end of the queue)
	if (fast) {
		dataStr (&(ring->buf[ring->end]), data);           // insert
		if (ring->size < 1) {
			cerr << "JobListDrv::schedule: ring too small (<1)!" << endl;
			throw runtime_error ("JobListDrv::schedule: ring too small (<1)!");
		}
		ring->end = (ring->end+1)%ring->size;    // set new 'end' position (one further)

		if ((ring->pos==ring->end-1) && ring->notempty)
			ring->pos=(ring->pos+1)%ring->size; // push 'pos' one further in case of overrun

		ring->notempty = true;
	} else {
		throw runtime_error ("JobListDrv::schedule: only fast method implemented!"
				     " Be sure you know what that means (rtfs).");
	}
	
	return true;
}

void JobListDrv::lock(void)
{
	semctl (semid, 0, SETVAL, 1);
}

void JobListDrv::unlock(void)
{
	semctl (semid, 0, SETVAL, 0);
}

//  void JobListDrv::readList (int fd)
//  {
//  	ifstream in (fd);
//  	char line[256];
//  	string mod("skydrv");
//  	bool first=true;
//  	int nr = 0;
	
//  	cerr << "reading from " << dec(fd) << endl;
//  	while (!(in.getline(line,255)).eof()) {
//  		//cerr << "line: <" << line << ">" << endl;
//  		if (!line[0] || line[0]==10 || line[0]==13) goto readok;
//  		else {
//  			// data format: %f %f %f\n
//  			char *l = line;
//  			char *time = line;
//  			double dtime = atof(l);

//  			while (*l && *l!=' ') { l++; };
//  			if (*l) *l=0; *l++;
//  			char *azm = l;
//  			double dazm = atof(azm);

//  			while (*l && *l!=' ') { l++; };
//  			if (*l) *l=0; *l++;
//  			char *elv = l;
//  			double delv = atof(elv);

//  			while (*l && *l!=' ') { l++; };
//  			if (*l) *l=0;
//  			l++;
//  			char *dst = l;
//  			double ddst = atof(dst);
			
//  			JobData data;
//  			data["time"] = time;
//  			data["azm"] = azm;
//  			data["elv"] = elv;
//  			data["dst"] = dst;
//  			data.setStart (dtime-1.0); // 1 second time. Drive-Timeslice !!
//  			data.setLen (0.0);
//  			data.setMod (mod);

//  			if (first) {
//  				// if this is the first entry in the list
//  				// we will need some more time to bring
//  				// the telescope to the right position
//  				Time ttsetup = 300.0; // 5 minutes
//  				Time oldt = data.start();

//  				data.setStart (oldt-ttsetup); // first move
//  				this->schedule (data);

//  				// further moves -- accel correction
//  				data.setStart (oldt-ttsetup/5.0);
//  				this->schedule (data);
//  				data.setStart (oldt-ttsetup/10.0);
//  				this->schedule (data);
//  				data.setStart (oldt-ttsetup/20.0);
//  				this->schedule (data);
				
//  				// final schedule -- this time it counts ;)
//  				data.setStart (oldt);
//  				this->schedule (data);
				
//  				first = false;
//  			} else {
//  				this->schedule (data);
//  			}
//  			nr++;
//  		}
//  	}
//   readok:
//  	cerr << "JobListDrv: added " << nr << " jobs (plus 4 init jobs)." << endl;
//  	return;
//  }

JobListDrv *factory_JobListDrv (void)
{
	cerr << "creating JobListDrv" << endl;
	return new JobListDrv;
}
