#include <sky/drv/jobshelldrv.hh>

#include <stdexcept>
#include <sys/ioctl.h>
#include <linux/stepeng_ioctl.h>
#include <fcntl.h>

#include <cfg/cfg.hh>

#define DEFKEY 3

JobShellDrv::JobShellDrv (const string &i)
	:dkey(DEFKEY),dd(NULL),action(ActMoveTo)
{
	if (i.length()) this->init (i);
}


JobShellDrv::~JobShellDrv ()
{
	this->cleanup();
}

void JobShellDrv::init (const string &i)
{
	map<string,string> cfg;
	strToMap (i, cfg);
	this->init(cfg);
}

void JobShellDrv::init (map<string,string> &cfg)
{
	dkey = atoi(cfg["ddkey"].c_str());
	dd = dd_init(dkey);
	if (cfg["action"] == "ActMoveTo") action = ActMoveTo;
	else if (cfg["action"] == "ActSetSpeed") action = ActSetSpeed;
}

void JobShellDrv::cleanup ()
{
	dd_close (dd);
}

/*
 * Theory and practice:
 * 
 * This is called every time the telescope position has to be updated.
 * That means at least once a second, so you shouldn't waste too much time.
 * The procedure assumes that the telescope is able to reach the given speed
 * in zero time. This however is only true for small speed _correction_, and
 * not for startup.
 * The driver -- if configured properly -- will set the given speed with
 * attention to the maximum acceleration. To be more precise, this means that
 * the telescope will need more time to reach the given speed, and thus the
 * given point (for example if it has to be massively moved first, from one side
 * to the other...) So consider schedunling the first move some time ahead the
 * second one (depending on the hardware (telescope, drive etc) few minutes should
 * do). The second problem is that we will not reach the requested point in time.
 * With iterations following, the new speed is calculated using the _actual_ speed
 * of the engine, and not the theoretical, so after 4 or 5 iterations, the
 * error ought to be corredted.
 *
 * To be clear, this means you have to schedule things the following way:
 *
 *      o 1 move to the start point (or close to the starting point)
 *        time: minutes.
 *      o few moves (<10?) to get in sync with the all-together move we have
 *        to follow
 *      o the all-together move we got from the sky-core (the planet route).
 */
bool JobShellDrv::exec (const JobData &data)
{
	if (action == ActMoveTo) dd_goto (dd, atof(data.fetch("time").c_str()), // <= the time (until when)
					  atof(data.fetch("x").c_str()),
					  atof(data.fetch("y").c_str()),
					  atof(data.fetch("z").c_str()));
	else if (action == ActSetSpeed) dd_setspeed (dd, atof(data.fetch("z").c_str()), // <= the time (how long)
						     atof(data.fetch("x").c_str()),
						     atof(data.fetch("y").c_str()));
}

/* the factory */
void *factory_JobShellDrv (void)
{
	return new JobShellDrv();
}
